﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Nyerguds.Ini;
using System.IO;
using System.Diagnostics;
using System.Text.RegularExpressions;

namespace war2_config
{
    public partial class MainForm : Form
    {
        readonly string CnCDDrawIniPath = Program.PATH + "ddraw.ini";
        readonly string War2DDrawIniPath = Program.PATH + "war2_ddraw.ini";
        readonly string DDrawDllPath = Program.PATH + "ddraw.dll";
        readonly string HardwareCursorPluginPath = Program.PATH + Path.Combine("plugin", "HardwareCursor.w2p");
        readonly string HardwareCursorIniPath = Program.PATH + Path.Combine("plugin", "HardwareCursor.ini");
        readonly string PluginPath = Program.PATH + "plugin" + Path.DirectorySeparatorChar;

        IniFile CnCDDrawIni, War2DDrawIni;

        struct Renderer { public string Name; public byte[] Data; };
        Renderer GDIRenderer         = new Renderer { Name = "GDI",        Data = Properties.Resources.war2_ddraw };
        Renderer OpenGLRenderer      = new Renderer { Name = "OpenGL",     Data = Properties.Resources.cnc_ddraw };
        Renderer Direct3D9Renderer   = new Renderer { Name = "Direct3D 9", Data = Properties.Resources.cnc_ddraw };
        Renderer DirectDrawRenderer  = new Renderer { Name = "DirectDraw", Data = new byte[0] };


        string FullScreenDisplayMode = "Full Screen";
        string WindowedDisplayMode   = "Windowed";
        string StretchedDisplayMode  = "Stretched";
        string BorderlessDisplayMode = "Borderless";

        string CurrentDisplayMode;


        public MainForm()
        {
            InitializeComponent();
            Program.Translation.ApplyTo(this, toolTip1);
            Program.Translation.ApplyTo(ref FullScreenDisplayMode);
            Program.Translation.ApplyTo(ref WindowedDisplayMode);
            Program.Translation.ApplyTo(ref StretchedDisplayMode);
            Program.Translation.ApplyTo(ref BorderlessDisplayMode);

            if (!File.Exists(CnCDDrawIniPath))
                File.WriteAllBytes(CnCDDrawIniPath, Properties.Resources.ddraw_ini);

            if (Directory.Exists(PluginPath) && !File.Exists(HardwareCursorIniPath))
                File.WriteAllBytes(HardwareCursorIniPath, Properties.Resources.HardwareCursor1);

            CnCDDrawIni = new IniFile(CnCDDrawIniPath, BooleanMode.TRUE_FALSE);

            if (!File.Exists(War2DDrawIniPath))
                File.WriteAllBytes(War2DDrawIniPath, Properties.Resources.war2_ddraw_ini);

            War2DDrawIni = new IniFile(War2DDrawIniPath, BooleanMode.YES_NO);

            RendererComboBox.Items.Add(GDIRenderer.Name);
            RendererComboBox.Items.Add(OpenGLRenderer.Name);
            RendererComboBox.Items.Add(Direct3D9Renderer.Name);
            RendererComboBox.Items.Add(DirectDrawRenderer.Name);

            DisplayModeComboBox.Items.Add(FullScreenDisplayMode);
            DisplayModeComboBox.Items.Add(WindowedDisplayMode);
            DisplayModeComboBox.Items.Add(StretchedDisplayMode);
            DisplayModeComboBox.Items.Add(BorderlessDisplayMode);

            HardwareCursorCheckBox.Visible = Directory.Exists(PluginPath);
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            try
            {
                var shaders = Directory.GetFiles(Program.PATH + "Shaders", "*.glsl");
                if (shaders != null && shaders.Length > 0)
                {
                    foreach (string shader in shaders)
                        FilterComboBox.Items.Add(Path.GetFileNameWithoutExtension(shader));
                }
            }
            catch { }

            LoadSettings();
        }

        private void OkButton_Click(object sender, EventArgs e)
        {
            SaveSettings();
            Application.Exit();
        }

        private void CancelButton_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void AdvancedSettingsLabel_Click(object sender, EventArgs e)
        {
            //SaveSettings();
            //Application.Exit();
        }

        private void RendererComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            bool usingCnCDDraw = 
                RendererComboBox.SelectedItem.Equals(Direct3D9Renderer.Name) ||
                RendererComboBox.SelectedItem.Equals(OpenGLRenderer.Name);

            DisplayModeLabel.Enabled = DisplayModeComboBox.Enabled = 
                !RendererComboBox.SelectedItem.Equals(DirectDrawRenderer.Name);

            MaintainAspectRatioCheckBox.Enabled = !RendererComboBox.SelectedItem.Equals(DirectDrawRenderer.Name);
            ScaleMouseCheckBox.Enabled = !RendererComboBox.SelectedItem.Equals(DirectDrawRenderer.Name) && !HardwareCursorCheckBox.Checked;
            FilterLabel.Enabled = FilterComboBox.Enabled = RendererComboBox.SelectedItem.Equals(OpenGLRenderer.Name);

            if (usingCnCDDraw)
            {
                if (DisplayModeComboBox.Items.Count != 4)
                {
                    DisplayModeComboBox.Items.Add(StretchedDisplayMode);
                    DisplayModeComboBox.Items.Add(BorderlessDisplayMode);
                }
            }
            else
            {
                if (DisplayModeComboBox.Items.Count == 4)
                {
                    DisplayModeComboBox.Items.Remove(StretchedDisplayMode);
                    DisplayModeComboBox.Items.Remove(BorderlessDisplayMode);
                }
            }

            if (DisplayModeComboBox.SelectedIndex == -1)
                DisplayModeComboBox.SelectedIndex = 0;
        }
        private void HardwareCursorCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            ScaleMouseCheckBox.Enabled = !RendererComboBox.SelectedItem.Equals(DirectDrawRenderer.Name) && !HardwareCursorCheckBox.Checked;
        }

        private void LoadSettings()
        {
            long ddrawSize = File.Exists(DDrawDllPath) ? new FileInfo(DDrawDllPath).Length : 0;

            if (ddrawSize == OpenGLRenderer.Data.Length)
            {
                if (CnCDDrawIni.GetStringValue("ddraw", "renderer", "auto") == "direct3d9")
                {
                    RendererComboBox.SelectedItem = Direct3D9Renderer.Name;
                }
                else
                {
                    RendererComboBox.SelectedItem = OpenGLRenderer.Name;
                }

                MaintainAspectRatioCheckBox.Checked = CnCDDrawIni.GetBoolValue("ddraw", "maintas", false);
                ScaleMouseCheckBox.Checked = CnCDDrawIni.GetBoolValue("ddraw", "adjmouse", false);
            }
            else if (ddrawSize > 0)
            {
                RendererComboBox.SelectedItem = GDIRenderer.Name;

                MaintainAspectRatioCheckBox.Checked = War2DDrawIni.GetBoolValue("ddraw", "MaintainAspectRatio", false);
                ScaleMouseCheckBox.Checked = War2DDrawIni.GetBoolValue("ddraw", "AdjustMouseSensitivity", false);
            }
            else
                RendererComboBox.SelectedItem = DirectDrawRenderer.Name;


            bool stretched = CnCDDrawIni.GetBoolValue("ddraw", "fullscreen", false);
            bool windowed = CnCDDrawIni.GetBoolValue("ddraw", "windowed", false);

            if (RendererComboBox.SelectedItem.Equals(DirectDrawRenderer.Name))
            {
                DisplayModeComboBox.SelectedItem = FullScreenDisplayMode;
            }
            else if (RendererComboBox.SelectedItem.Equals(GDIRenderer.Name))
            {
                if (War2DDrawIni.GetBoolValue("ddraw", "FullscreenFailed", false))
                    DisplayModeComboBox.SelectedItem = FullScreenDisplayMode;
                else if (War2DDrawIni.GetBoolValue("ddraw", "Windowed", false))
                    DisplayModeComboBox.SelectedItem = WindowedDisplayMode;
                else
                    DisplayModeComboBox.SelectedItem = FullScreenDisplayMode;
            }
            else if (windowed && stretched)
            {
                DisplayModeComboBox.SelectedItem = BorderlessDisplayMode;
            }
            else if (windowed)
            {
                DisplayModeComboBox.SelectedItem = WindowedDisplayMode;
            }
            else if (stretched)
            {
                DisplayModeComboBox.SelectedItem = StretchedDisplayMode;
            }
            else
                DisplayModeComboBox.SelectedItem = FullScreenDisplayMode;

            CurrentDisplayMode = DisplayModeComboBox.SelectedItem.ToString();

            string filter = Path.GetFileNameWithoutExtension(CnCDDrawIni.GetStringValue("ddraw", "shader", ""));

            if (FilterComboBox.Items.Contains(filter))
                FilterComboBox.SelectedItem = filter;
            else
                FilterComboBox.SelectedIndex = 0; //nearest

            HardwareCursorCheckBox.Checked = File.Exists(HardwareCursorPluginPath);
        }

        private void SaveSettings()
        {
            if (HardwareCursorCheckBox.Checked)
                ScaleMouseCheckBox.Checked = true;

            if (RendererComboBox.SelectedItem.Equals(Direct3D9Renderer.Name) ||
                RendererComboBox.SelectedItem.Equals(OpenGLRenderer.Name))
            {
                string renderer = RendererComboBox.SelectedItem.Equals(Direct3D9Renderer.Name) ? "direct3d9" : "opengl";

                CnCDDrawIni.SetStringValue("ddraw", "renderer", renderer);

                bool windowed = 
                    DisplayModeComboBox.SelectedItem.Equals(WindowedDisplayMode) ||
                    DisplayModeComboBox.SelectedItem.Equals(BorderlessDisplayMode);

                CnCDDrawIni.SetBoolValue("ddraw", "windowed", windowed);

                bool stretched =
                    DisplayModeComboBox.SelectedItem.Equals(StretchedDisplayMode) ||
                    DisplayModeComboBox.SelectedItem.Equals(BorderlessDisplayMode);

                CnCDDrawIni.SetBoolValue("ddraw", "fullscreen", stretched);

                CnCDDrawIni.SetBoolValue("ddraw", "maintas", MaintainAspectRatioCheckBox.Checked);
                CnCDDrawIni.SetBoolValue("ddraw", "adjmouse", ScaleMouseCheckBox.Checked);

                if (FilterComboBox.SelectedIndex == 0)
                {
                    CnCDDrawIni.SetStringValue("ddraw", "shader", "");
                }
                else
                {
                    CnCDDrawIni.SetStringValue(
                        "ddraw", 
                        "shader",
                        "Shaders" + Path.DirectorySeparatorChar + FilterComboBox.SelectedItem.ToString() + ".glsl");
                }
            }
            else if (RendererComboBox.SelectedItem.Equals(GDIRenderer.Name))
            {
                if (War2DDrawIni.GetBoolValue("ddraw", "FullscreenFailed", false))
                {
                    if (!DisplayModeComboBox.SelectedItem.Equals(FullScreenDisplayMode))
                    {
                        War2DDrawIni.SetBoolValue("ddraw", "ShowWindowFrame", true);
                        War2DDrawIni.SetBoolValue("ddraw", "FullscreenFailed", false);
                        War2DDrawIni.SetIntValue("ddraw", "PosX", -32000);
                        War2DDrawIni.SetIntValue("ddraw", "PosY", -32000);
                    }
                }
                else
                {
                    War2DDrawIni.SetBoolValue(
                        "ddraw",
                        "Windowed",
                        DisplayModeComboBox.SelectedItem.Equals(WindowedDisplayMode));
                }

                War2DDrawIni.SetBoolValue("ddraw", "MaintainAspectRatio", MaintainAspectRatioCheckBox.Checked);
                War2DDrawIni.SetBoolValue("ddraw", "AdjustMouseSensitivity", ScaleMouseCheckBox.Checked);
            }

            if (File.Exists(HardwareCursorPluginPath))
            {
                if (!HardwareCursorCheckBox.Checked)
                    File.Delete(HardwareCursorPluginPath);
            }
            else if (HardwareCursorCheckBox.Checked)
            {
                File.WriteAllBytes(HardwareCursorPluginPath, Properties.Resources.HardwareCursor);
            }

            if (DisplayModeComboBox.SelectedItem.ToString() != CurrentDisplayMode)
            {
                War2DDrawIni.SetIntValue("ddraw", "width", 640);
                War2DDrawIni.SetIntValue("ddraw", "height", 480);
                CnCDDrawIni.SetIntValue("ddraw", "width", 0);
                CnCDDrawIni.SetIntValue("ddraw", "height", 0);
            }

            CnCDDrawIni.WriteIni();
            War2DDrawIni.WriteIni();

            if (RendererComboBox.SelectedItem.Equals(GDIRenderer.Name))
            {
                File.WriteAllBytes(DDrawDllPath, GDIRenderer.Data);
            }
            else if (
                RendererComboBox.SelectedItem.Equals(Direct3D9Renderer.Name) ||
                RendererComboBox.SelectedItem.Equals(OpenGLRenderer.Name))
            {
                File.WriteAllBytes(DDrawDllPath, Direct3D9Renderer.Data);
            }
            else
                File.Delete(DDrawDllPath);
        }

    }
}
